/* 
   hash table handler 
  
  hashes pointers to an object with a string index
 note, uses (void *) ptr to be able to use any type.

 to use it you probably define some macros that cast the pointers to 
 the right type

note:
  objects are not freed when hashtable is destroyed

*/


#include <stdlib.h>
#include <string.h>

#include <error.h>
#include <dbgmem.h>

#include <hash.h>

#define HASHGROW 10

static unsigned int HashFunc(char *str, unsigned int size);
static void HashFreeTable(hashtable ht);
static void HashResize(hashtable ht);

hashtable HashCreate(int startsize)
{
  hashtable ht;
  int i;

  if (startsize<10) /* May bug if too small, unusable then anyway */
    startsize=10;

  ht=malloc(sizeof(struct hashtable));
  if (ht==0) {
    ErrError("Out of memory creating hashtable");
  }

  ht->size=startsize;

  ht->numelements=0;
  ht->table=malloc(ht->size*sizeof(hashobject));
  if (ht->table==0) {
    ErrError("Out of memory creating hashtable");
  }

  for(i=0;i<ht->size;i++) {
    ht->table[i].string=0;
    ht->table[i].object=0;
  }

  return ht;
}

static void HashFreeTable(hashtable ht)
{
  int i;

  if (ht->table!=0) {
    for(i=0;i<ht->size;i++) {
      if (ht->table[i].string!=0) {
	free(ht->table[i].string);
      }
    }
    free(ht->table);
  }
}   


void HashDestroy(hashtable ht)
{
  if (ht!=0) {
    HashFreeTable(ht);    
    free(ht);
  }
}   
    
  
void HashInsert(hashtable ht, void *object, char *string)
{
  unsigned int h;

  /* make sure table is not full! (max 80 %) */
  if ( (ht->numelements*100)/ht->size > 80) { 
    HashResize(ht);
  }

  h=HashFunc(string,ht->size);
  
  /* guaranteed to find one, because table is not full */
  while((ht->table[h].string!=0) && (strcmp(string,ht->table[h].string)!=0)){ 
    h=(h+1)%ht->size;
  }

  ht->table[h].string=malloc(sizeof(char)*(strlen(string)+1));
  if (ht->table[h].string==0) {
    ErrError("Out of memory inserting object in hashtable.");
  }
  strcpy(ht->table[h].string,string);
  ht->table[h].object=object;

  ht->numelements++;
}

static void HashResize(hashtable ht)
{
  hashtable newht;
  int i;
  
  newht=HashCreate(ht->size+HASHGROW);
  for (i=0;i<ht->size;i++) {
    if (ht->table[i].string!=0) {
      HashInsert(newht, ht->table[i].object, ht->table[i].string);
    }
  }

  HashFreeTable(ht);
  *ht=*newht;
  free(newht);
}

void *HashLookup(hashtable ht, char *string)
{
  unsigned int h;

 
  h=HashFunc(string,ht->size);
  
  /* guaranteed not to wrap, because table is not full */
  while((ht->table[h].string!=0) && (strcmp(string,ht->table[h].string)!=0)) { 
    h=(h+1)%ht->size;
  }
  
  if (ht->table[h].string==0) {
    return 0;
  }

  return ht->table[h].object;
}

void HashApply(hashtable ht, void (*fun)(char *str,void *obj))
{
  int i;

  for (i=0;i<ht->size;i++) {
    if (ht->table[i].string!=0) {
      fun( ht->table[i].string, ht->table[i].object );
    }
  }
}

void HashDelete(hashtable ht, char *string)
{
  unsigned int h;

 
  h=HashFunc(string,ht->size);
  
  /* guaranteed not to wrap, because table is not full */
  while((ht->table[h].string!=0) && (strcmp(string,ht->table[h].string)!=0)) { 
    h=(h+1)%ht->size;
  }
  
  if (ht->table[h].string==0) {
    return;
  }
  
  free(ht->table[h].string);
  ht->table[h].string=0;
  ht->table[h].object=0;
}


static unsigned int HashFunc(char *s, unsigned int size)
{
  unsigned hashval;

  for (hashval=0; *s!=0; s++) {
    hashval= *s +31*hashval;
  }  

  return hashval%size;
}

